# build.py

import os
import codecs
import pathlib
from setuptools import setup as _setup, find_namespace_packages
from typing import Iterable, Any

__all__ = [
    "Dependencies",
    "parse_requirements",
    "unfold",
    "get_dependencies",
    "build_manifest",
    "build_pyproject",
    "collect_files",
    "setup"
]

class Dependencies:
    """A class to represent dependencies."""

    def __init__(
            self,
            requirements: Iterable[str] = None,
            links: Iterable[str] = None
    ) -> None:
        """
        Defines the class attributes.

        :param requirements: The requirement names and settings.
        :param links: The requirement links.
        """

        if requirements is None:
            requirements = []
        # end if

        if links is None:
            links = []
        # end if

        self.requirements = requirements
        self.links = links
    # end __init__
# end Dependencies

def parse_requirements(
        source: str | pathlib.Path,
        excluded: Iterable[str] = None
) -> Dependencies:
    """
    Parses the requirements file.

    :param source: The source of the requirements.
    :param excluded: Names of dependencies to exclude.

    :return: A dependencies object.
    """

    if excluded is None:
        excluded = []
    # end if

    links = []
    requirements = []

    with codecs.open(source, 'r') as requirements_txt:
        for line in requirements_txt.readlines():
            line = line.replace("\n", "")

            if not line or line.replace(" ", "").startswith("#"):
                continue
            # end if

            line = line[:line.find("#")] if "#" in line else line

            requirement = line.split(";")

            for i, part in enumerate(parts := requirement[0].split()):
                name = parts[0]

                for char in ['=', '<', ">"]:
                    name = name[:name.find(char)] if char in name else name
                # end for

                if name in excluded:
                    continue
                # end if

                if (
                    (len(parts) > 2) and
                    (part in ("--extra-index-url", "--index-url", "--find-links"))
                ):
                    if i == 0:
                        name = parts[-1]

                    else:
                        name = parts[0]
                    # end if

                    if name in excluded:
                        continue
                    # end if

                    links.append(f"{parts[1]}#egg={name}")
                    requirements.append(";".join([name] + requirement[1:]))

                    break
                # end if

                if len(parts) == 1 and name not in excluded:
                    requirements.append(line)
                # end if
            # end for
        # end for
    # end open

    return Dependencies(
        requirements=requirements,
        links=links
    )
# parse_requirements

def unfold(obj: Any, /, indentation: int = None) -> str:
    """
    Unfolds the object if it is an iterable.

    :param obj: The object to convert to a string.
    :param indentation: The indentation level.

    :return: The string representation of the object.
    """

    try:
        if isinstance(obj, str):
            raise TypeError
        # end if

        string = str(obj)

        if not string:
            raise TypeError
        # end if

        if indentation is None:
            indentation = 1
        # end if

        return (
                string[0] + "\n" + ("\t" * indentation) +
                (
                    (",\n" + ("\t" * indentation)).join(
                        (
                            unfold(element, indentation=indentation + 1)
                            for element in iter(obj)
                        )
                        if not isinstance(obj, dict) else
                        (
                            (
                                f"{unfold(key, indentation=indentation + 1)}: "
                                f"{unfold(value, indentation=indentation + 1)}"
                            )
                            for key, value in obj.items()
                        )
                    )
                ) + "\n" + ("\t" * (indentation - 1)) +
                string[-1]
        )

    except TypeError:
        if isinstance(obj, bool):
            return repr(obj).lower()
        # end if

        return repr(obj)
    # end try
# end unfold

def get_dependencies(
        *,
        requirements: str | pathlib.Path = None,
        dev_requirements: str | pathlib.Path = None
) -> tuple[Dependencies, Dependencies]:
    """
    Runs the program to create the package.

    :param requirements: The requirement names and settings.
    :param dev_requirements: The requirement names and settings.
    """

    if requirements is not None and os.path.exists(requirements):
        dependencies = parse_requirements(source=requirements)

        if dev_requirements is not None and os.path.exists(dev_requirements):
            extra_dependencies = parse_requirements(source=dev_requirements)

            with codecs.open(requirements, 'r') as req_file:
                reqs = req_file.read()
            # end open

            with codecs.open(dev_requirements, 'r') as req_dev_file:
                dev_reqs = req_dev_file.read()
            # end open

            if reqs not in dev_reqs:
                with codecs.open(dev_requirements, 'a') as req_dev_file:
                    req_dev_file.write(f"\n\n{reqs}")
                # end open
            # end if

        else:
            extra_dependencies = Dependencies()
        # end if

    else:
        dependencies = Dependencies()
        extra_dependencies = Dependencies()
    # end if

    return dependencies, extra_dependencies
# end get_dependencies

def build_pyproject(
        project: str | pathlib.Path = None, **kwargs: Any
) -> None:
    """
    Builds the package.

    :param project: The project configuration toml file.
    :param kwargs: Any keyword argument.
    """

    if project is not None:
        content = "\n".join(
            [
                f"{key} = {unfold(value, indentation=1)}"
                for key, value in kwargs.items() if not (
                    isinstance(value, dict) or
                    (
                        key in (
                            "long_description",
                            "license",
                            'include_package_data',
                            'dependency_links',
                            'url',
                            'install_requires',
                            'extras_require',
                            'author',
                            'packages',
                            'author_email',
                            "data_files",
                            "long_description_content_type"
                        )
                    )
                )
            ]
        )

        if os.path.exists(project):
            with codecs.open(project, 'r') as project_file:
                project_content = project_file.read()
            # end open

        else:
            project_content = ""
        # end if

        if content not in project_content:
            with codecs.open(project, 'w') as req_dev_file:
                req_dev_file.write(
                    '[build-system]\n'
                    'requires = ["setuptools"]\n'
                    'build-backend = "setuptools.build_meta"\n'
                    f'\n[project]\n{content}'
                )
            # end open
        # end if
    # end if
# end build_pyproject

def build_manifest(
        include: Iterable[str] = None, manifest: bool = None
) -> None:
    """
    Builds the manifest file.

    :param include: The files to include.
    :param manifest: The value to read the content from the file.
    """

    if not os.path.exists("MANIFEST.in") or not manifest:
        with codecs.open("MANIFEST.in", 'w') as include_file:
            include_file.write("")
        # end open
    # end if

    if os.path.exists("MANIFEST.in"):
        with codecs.open("MANIFEST.in", 'r') as include_file:
            include_content = include_file.read()
        # end open

        with codecs.open("MANIFEST.in", 'a') as include_file:
            for line in include:
                if line not in include_content:
                    include_file.write(f"include {line}\n")
                # end if
            # end for
        # end open
    # end if
# end build_manifest
def collect_files(location: str | pathlib.Path, levels: int = None) -> list[str]:
    """
    Collects all the file paths from the location with the extension.

    :param location: The location of the files.
    :param levels: The search levels.

    :return: A list of file paths.
    """

    paths = []

    if levels == 0:
        return paths
    # end if

    for name in os.listdir(location):
        path = pathlib.Path(location) / pathlib.Path(name)

        if path.is_file():
            paths.append(str(path))

        elif path.is_dir():
            paths.extend(
                collect_files(
                    str(path), levels=(
                        levels - 1 if levels is not None else levels
                    )
                )
            )
        # end if
    # end for

    return paths
# end collect_files

def setup(
        package: str | pathlib.Path, *,
        readme: str | bool | pathlib.Path = None,
        exclude: Iterable[str | pathlib.Path] = None,
        include: Iterable[str | pathlib.Path] = None,
        requirements: str | pathlib.Path = None,
        dev_requirements: str | pathlib.Path = None,
        project: str | pathlib.Path = None,
        manifest: bool = None,
        **kwargs: Any
) -> None:
    """
    Runs the program to create the package.

    :param package: The package path.
    :param readme :The path to the readme file.
    :param exclude: The paths to exclude.
    :param include: The paths to include.
    :param requirements: The requirement names and settings.
    :param dev_requirements: The requirement names and settings.
    :param project: The project configuration toml file.
    :param manifest: The value to read the content from the file.
    :param kwargs: The setup kwargs.
    """

    if include is None:
        include = []

    else:
        include = list(include)
    # end if

    if readme in (None, True):
        readme = 'README.md' if os.path.exists('README.md') else None
    # end if

    if exclude is None:
        exclude = ()
    # end if

    if isinstance(readme, (str, pathlib.Path)):
        with codecs.open(str(readme), 'r') as desc_file:
            long_description = desc_file.read()
        # end open

    else:
        long_description = None
    # end if

    dependencies, extra_dependencies = get_dependencies(
        requirements=requirements, dev_requirements=dev_requirements
    )

    include = list(
        set(
            include + [
                file for file in
                [requirements, dev_requirements, project, "build.py"]
                if file and os.path.exists(str(file))
            ]
        )
    )

    for included in list(include):
        if included is not None and os.path.isdir(str(included)):
            include += list(set(collect_files(location=included)))
        # end if
    # end for

    include = [str(pathlib.Path(path)) for path in include]

    packages = [
        f"{'.'.join(pathlib.Path(str(package)).parts)}.{content}"
        for content in find_namespace_packages(package, exclude=exclude)
    ] + [package]

    kwargs.setdefault("extras_require", {}).setdefault(
        "dev", [
            req for req in extra_dependencies.requirements
            if req not in dependencies.requirements
        ]
    )

    build_manifest(include=include, manifest=manifest)

    if project:
        build_pyproject(project=project, **kwargs)
    # end if

    _setup(
        **(dict(project=project) if project else {}),
        packages=packages,
        long_description=long_description,
        include_package_data=True,
        dependency_links=dependencies.links,
        install_requires=dependencies.requirements,
        **kwargs
    )
# end setup


# queue.py

import datetime as dt
from typing import Callable, Any
import socket as _socket

from looperator import Operator, Handler

from socketsio.sockets import Socket

__all__ = [
    "SocketSenderQueue"
]

Connection = _socket.socket
Address = tuple[str, int]

class SocketSenderQueue(Operator):
    """
    A class to schedule and manage sending and receiving
    messages across multiple processes.
    """

    def __init__(
            self,
            socket: Socket,
            queue: list[tuple[bytes, Address | None]] = None,
            termination: Callable[[], Any] = None,
            handler: Handler = None,
            loop: bool = True,
            delay: float | dt.timedelta = None,
            block: bool = False,
            wait: float | dt.timedelta | dt.datetime = None,
            timeout: float | dt.timedelta | dt.datetime = None
    ) -> None:
        """
        Defines the attributes of the socket sender queue.

        :param socket: The base socket object.
        :param queue: The queue object.
        :param termination: The termination callback.
        :param handler: The handler object to handle the operation.
        :param loop: The value to run a loop.
        :param delay: The delay for the process.
        :param wait: The value to wait after starting to run the process.
        :param block: The value to block the execution.
        :param timeout: The valur to add a start_timeout to the process.
        """

        self.socket = socket

        self.queue = queue or []

        super().__init__(
            operation=self.send_queue,
            termination=termination,
            handler=handler,
            loop=loop,
            delay=delay,
            block=block,
            wait=wait,
            timeout=timeout
        )
    # end __init__

    @property
    def address(self) -> Address:
        """
        Returns the ip and port of the binding.

        :return: The address tuple.
        """

        return self.socket.address
    # end address

    def is_tcp(self) -> bool:
        """
        Checks if the socket is a TCP socket.

        :return: The boolean flag.
        """

        return self.socket.is_tcp()
    # end is_tcp

    def is_udp(self) -> bool:
        """
        Checks if the socket is a UDP socket.

        :return: The boolean flag.
        """

        return self.socket.is_udp()
    # end is_udp

    def is_tcp_bluetooth(self) -> bool:
        """
        Checks if the socket is a UDP socket.

        :return: The boolean flag.
        """

        return self.socket.is_tcp_bluetooth()
    # end is_tcp_bluetooth

    def validate_connection(self) -> None:
        """Validates a connection."""

        self.socket.validate_connection()
    # end validate_connection

    def send_queue(self) -> None:
        """Sends the message from the queue"""

        if self.queue:
            try:
                data = self.queue.pop(0)

                if data[0]:
                    print(data)
                    self.socket.send(*data)
                # end if

            except IndexError:
                pass
            # end try
        # end if
    # end send_queue

    def send_all_queue(self) -> None:
        """Sends the message from the queue"""

        while self.queue:
            self.send_queue()
        # end while
    # end send_all_queue

    def receive(self, address: Address = None) -> tuple[bytes, Address | None]:
        """
        Receive a message from the client or server by its connection.

        :param address: The address of the sender.

        :return: The received message from the server.
        """

        return self.socket.receive(address)
    # end receive

    def send(self, data: bytes, address: Address = None) -> tuple[bytes, Address | None]:
        """
        Sends a message to the client or server by its connection.

        :param data: The message to send to the client.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        self.queue.append((data, address))

        return data, address
    # end send

    def close(self) -> None:
        """Closes the connection."""

        self.socket.close()
    # end close
# end SocketQueue


# sockets.py

from typing import Self
import socket

from socketsio.protocols import (
    BaseProtocol, udp_socket, tcp_socket, is_udp,
    tcp_bluetooth_socket, is_tcp, is_tcp_bluetooth, is_udp_bluetooth
)

from represent import represent

__all__ = [
    "Socket",
    "tcp_socket",
    "udp_socket",
    "tcp_bluetooth_socket"
]

Connection = socket.socket
Address = tuple[str, int]
Output = tuple[bytes, Address | None]

@represent
class Socket:
    """A socket connection I/O object."""

    def __init__(
            self,
            protocol: BaseProtocol,
            connection: Connection = None,
            address: Address = None,
            reusable: bool = False
    ) -> None:
        """
        Defines the attributes of a server.

        :param connection: The socket object of the server.
        :param protocol: The communication protocol object.
        :param address: The address of the connection.
        :param reusable: The value to make the socket reusable.
        """

        self.connection = connection or protocol.socket()
        self.protocol = protocol
        self._reusable = reusable
        self._address = address

        self._closed = False

        self._connections_count = 1
    # end __init__

    @property
    def empty(self) -> bool:
        """
        Checks if the socket is empty.

        :return: The value of the connection.
        """

        return self.connection is None
    # end empty

    @property
    def closed(self) -> bool:
        """
        Checks if the socket is empty.

        :return: The value of the connection.
        """

        return self._closed
    # end closed

    @property
    def usable(self) -> bool:
        """
        Checks if the socket is usable.

        :return: The value of the connection.
        """

        return not self.empty or self.reusable
    # end usable

    @property
    def reusable(self) -> bool:
        """
        Checks if the socket is usable.

        :return: The value of the connection.
        """

        return self._reusable
    # end reusable

    @property
    def address(self) -> Address:
        """
        Returns the ip and port of the binding.

        :return: The address tuple.
        """

        return self._address
    # end address

    @property
    def connections_count(self) -> int:
        """
        Returns the connections count.

        :return: The count of connections made in the socket.
        """

        return self._connections_count
    # end connections_count

    def is_tcp(self) -> bool:
        """
        Checks if the socket is a TCP socket.

        :return: The boolean flag.
        """

        return is_tcp(self.connection or self.protocol.socket())
    # end is_tcp

    def is_udp(self) -> bool:
        """
        Checks if the socket is a UDP socket.

        :return: The boolean flag.
        """

        return is_udp(self.connection or self.protocol.socket())
    # end is_udp

    def is_tcp_bluetooth(self) -> bool:
        """
        Checks if the socket is a UDP socket.

        :return: The boolean flag.
        """

        return is_tcp_bluetooth(self.connection or self.protocol.socket())
    # end is_tcp_bluetooth

    def is_udp_bluetooth(self) -> bool:
        """
        Checks if the socket is a UDP socket.

        :return: The boolean flag.
        """

        return is_udp_bluetooth(self.connection or self.protocol.socket())
    # end is_udp_bluetooth

    def validate_open(self) -> None:
        """Validates that the socket is not closed."""

        if self.closed:
            raise ValueError(
                "Unable to operate using "
                f"{'an empty' if self.reusable else  'a used and non-reusable'}, "
                "server socket."
            )
        # end if
    # end validate_open

    def validate_connection(self) -> None:
        """Validates a connection."""

        if self.connection is None:
            if self.reusable:
                self.connection = self.protocol.socket()

                self._closed = False

                self._connections_count += 1

            elif self.connections_count > 0:
                raise ValueError(
                    "Only one connection is allowed on a non-reusable socket. "
                    "Consider making the socket reusable."
                )
            # end if
        # end if

        self.validate_open()
    # end validate_connection

    def send(
            self,
            data: bytes,
            connection: Connection = None,
            address: Address = None
    ) -> Output:
        """
        Sends a message to the client or server by its connection.

        :param data: The message to send to the client.
        :param connection: The sockets' connection object.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        return self.protocol.send(
            connection=connection or self.connection,
            data=data, address=address or self.address
        )
    # end send

    def receive(
            self,
            connection: Connection = None,
            address: Address = None
    ) -> Output:
        """
        Receive a message from the client or server by its connection.

        :param connection: The sockets' connection object.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        data, new_address = self.protocol.receive(
            connection=connection or self.connection,
            address=address or self.address
        )

        if address is None and self.address is None:
            self._address = new_address
        # end if

        return data, new_address
    # end receive

    def close(self) -> None:
        """Closes the connection."""

        self.connection.close()

        if self.reusable:
            self.connection = self.protocol.socket()

        else:
            self.connection = None

            self._closed = True
        # end if
    # end close

    def clone(self) -> Self:
        """
        Returns a copy of the socket wrapper object.

        :return: The new socket object.
        """

        return Socket(
            protocol=self.protocol,
            connection=self.protocol.socket(),
            address=self.address,
            reusable=self.reusable
        )
    # end clone
# end Socket


# __init__.py

from socketsio.sockets import *
from socketsio.server import *
from socketsio.client import *
from socketsio.protocols import *
from socketsio.utils import *
from socketsio.queue import SocketSenderQueue


# server.py

import socket
from typing import Any, Callable, Self
import threading

from socketsio.protocols import BaseProtocol
from socketsio.sockets import Socket

__all__ = [
    "Server",
    "ServerSideClient"
]

Connection = socket.socket
Address = tuple[str, int]
Action = Callable[[Socket], Any]
Output = tuple[bytes, Address | None]

class ServerSideClient(Socket):
    """A socket connection I/O object."""

    def __init__(
            self,
            protocol: BaseProtocol,
            connection: Connection = None,
            address: Address = None
    ) -> None:
        """
        Defines the attributes of a server.

        :param connection: The socket object of the server.
        :param protocol: The communication protocol object.
        :param address: The address of the connection.
        """

        super().__init__(
            protocol=protocol,
            connection=connection,
            address=address,
            reusable=False
        )

        self._connected = self.connection is not None
    # end __init__

    @property
    def connected(self) -> bool:
        """
        Returns the value of the socket being connected.

        :return: The connected value.
        """

        return self._connected
    # end connected

    def close(self) -> None:
        """Closes the connection."""

        if not self.connected or not self.connection:
            return
        # end if

        if not self.is_udp():
            self.connection.close()

            self.connection = None
        # end if

        self._connected = False
    # end close

    def make_reusable(self) -> None:
        """Makes the socket reusable."""

        self._reusable = True
    # end make_reusable

    def make_unreusable(self) -> None:
        """Makes the socket unreusable."""

        self._reusable = False
    # end make_unreusable

    def validate_connection(self) -> None:
        """Validates a connection."""

        if not self.connected or not self.connection:
            raise ValueError("The socket is already closed.")
        # end if
    # end validate_connection
# end ServerSideClient

class Server(Socket):
    """A class to represent the server object."""

    DELAY = 0.0001

    def __init__(
            self,
            protocol: BaseProtocol = None,
            connection: Connection = None,
            address: Address = None,
            reusable: bool = False,
            sequential: bool = True,
    ) -> None:
        """
        Defines the attributes of a server.

        :param connection: The socket object of the server.
        :param protocol: The communication protocol object.
        :param sequential: The value to sequentially find clients.
        :param address: The address to save for the socket.
        :param reusable: The value to make the socket reusable.
        """

        if sequential is None:
            sequential = False
        # end if

        super().__init__(
            connection=connection,
            protocol=protocol or BaseProtocol.protocol(),
            address=address,
            reusable=reusable
        )

        self._listening = False
        self._bound = False

        self.sequential = sequential
    # end __init__

    @property
    def listening(self) -> bool:
        """
        Returns the value of an active listening process.

        :return: The boolean flag.
        """

        return self._listening
    # end listening

    @property
    def bound(self) -> bool:
        """
        Returns the value of a bounded connection.

        :return: The boolean flag.
        """

        return self._bound
    # end bound

    @property
    def prebound(self) -> bool:
        """
        Returns the value of a bounded connection.

        :return: The boolean flag.
        """

        return self._address is not None
    # end prebound

    def make_reusable(self) -> None:
        """Makes the socket reusable."""

        self._reusable = True
    # end make_reusable

    def make_unreusable(self) -> None:
        """Makes the socket unreusable."""

        self._reusable = False
    # end make_unreusable

    def bind(self, address: Address) -> None:
        """
        Binds the connection of the server.

        :param address: The address to bind to.
        """

        self.validate_connection()

        self.connection.bind(address)

        self._address = address

        self._bound = True
    # end bind

    def prebind(self, address: Address) -> None:
        """
        Binds the connection of the server.

        :param address: The address to bind to.
        """

        self._address = address
    # end bind

    def rebind(self) -> None:
        """Binds the connection of the server."""

        self.bind(self._address)
    # end bind

    def validate_binding(self) -> None:
        """Validates the binding of the socket."""

        self.validate_connection()

        if not self.bound:
            if self.prebound:
                self.rebind()

            else:
                raise ValueError(
                    "Cannot start listening before binding."
                )
            # end if
        # end if
    # end validate_binding

    def validate_listening(self) -> None:
        """Validates the binding of the socket."""

        if not self.listening and not self.is_udp():
            self.listen()
        # end if
    # end validate_binding

    def accept(self) -> tuple[socket.socket, Address]:
        """
        Returns the connection and address from the accepted client.

        :return: The sockets object and the address.
        """

        self.validate_listening()

        return self.connection.accept()
    # end accept

    def _action_parameters(self, protocol: BaseProtocol = None) -> ServerSideClient:
        """
        Returns the parameters to call the action function.

        :return: The action parameters.
        """

        protocol = protocol or self.protocol

        if self.is_udp():
            connection, address = self.connection, None

        else:
            connection, address = self.accept()
        # end if

        return ServerSideClient(
            connection=connection,
            protocol=protocol,
            address=address
        )
    # end _action_parameters

    def send(
            self,
            data: bytes,
            connection: Connection = None,
            address: Address = None
    ) -> Output:
        """
        Sends a message to the client or server by its connection.

        :param connection: The sockets' connection object.
        :param data: The message to send to the client.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        if not self.is_udp():
            raise ValueError(
                "Cannot directly send/receive "
                "with a non-UDP server socket."
            )
        # end if

        self.validate_binding()

        return self.protocol.send(
            connection=connection or self.connection,
            data=data, address=address or self._address
        )
    # end send

    def receive(
            self,
            connection: Connection = None,
            address: Address = None
    ) -> Output:
        """
        Receive a message from the client or server by its connection.

        :param connection: The sockets' connection object.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        if not self.is_udp():
            raise ValueError(
                "Cannot directly send/receive "
                "with a non-UDP server socket."
            )
        # end if

        self.validate_binding()

        return self.protocol.receive(
            connection=connection or self.connection,
            address=address or self._address
        )
    # end receive

    def close(self) -> None:
        """Closes the connection."""

        super().close()

        self._listening = False
        self._bound = False
    # end close

    def action(self, client: Socket) -> None:
        """
        Sets or updates clients data in the clients' container .

        :param client: The client socket object.
        """
    # end action

    def listen(self) -> None:
        """Listens to clients."""

        self.validate_binding()

        self.connection.listen()

        self._listening = True
    # end listen

    def handle(
            self,
            protocol: BaseProtocol = None,
            action: Action = None,
            sequential: bool = None
    ) -> None:
        """
        Sends a message to the client by its connection.

        :param protocol: The protocol to use for sockets communication.
        :param action: The action to call.
        :param sequential: The value to sequentially find clients.
        """

        self.validate_listening()

        if sequential is not None:
            self.sequential = sequential
        # end if

        sequential = self.sequential

        if sequential:
            parameters = self._action_parameters(protocol=protocol)

            threading.Thread(target=lambda: action(parameters)).start()

        else:
            threading.Thread(
                target=lambda: action(
                    self._action_parameters(protocol=protocol)
                )
            ).start()
        # end if
    # end handle

    def clone(self) -> Self:
        """
        Returns a copy of the socket wrapper object.

        :return: The new socket object.
        """

        return Server(
            protocol=self.protocol,
            connection=self.protocol.socket(),
            address=self.address,
            reusable=self.reusable,
            sequential=self.sequential
        )
    # end clone
# end Server


# client.py

from typing import Self
import socket

from socketsio.protocols import BaseProtocol
from socketsio.sockets import Socket

__all__ = [
    "Client"
]

Connection = socket.socket
Address = tuple[str, int]
Output = tuple[bytes, Address | None]

class Client(Socket):
    """A class to represent the server object."""

    def __init__(
            self,
            protocol: BaseProtocol = None,
            connection: Connection = None,
            address: Address = None,
            reusable: bool = False
    ) -> None:
        """
        Defines the attributes of a server.

        :param connection: The socket object of the server.
        :param protocol: The communication protocol object.
        :param address: The address to save for the socket.
        :param reusable: The value to make the socket reusable.
        """

        super().__init__(
            connection=connection,
            protocol=protocol or BaseProtocol.protocol(),
            address=address,
            reusable=reusable
        )

        self._connected = False
    # end __init__

    @property
    def connected(self) -> bool:
        """
        Returns the value of a bounded connection.

        :return: The boolean flag.
        """

        return self._connected
    # end connected

    @property
    def preconnected(self) -> bool:
        """
        Returns the value of a bounded connection.

        :return: The boolean flag.
        """

        return self._address is not None
    # end preconnected

    def make_reusable(self) -> None:
        """Makes the socket reusable."""

        self._reusable = True
    # end make_reusable

    def make_unreusable(self) -> None:
        """Makes the socket unreusable."""

        self._reusable = False
    # end make_unreusable

    def connect(self, address: Address) -> None:
        """
        Returns the connection and address from the accepted client.

        :param address: The address of the server to connect to.
        """

        self.validate_connection()

        self.connection.connect(address)

        self._address = address

        self._connected = True
    # end connect

    def preconnect(self, address: Address) -> None:
        """
        Returns the connection and address from the accepted client.

        :param address: The address of the server to connect to.
        """

        self._address = address
    # end connect

    def reconnect(self) -> None:
        """Returns the connection and address from the accepted client."""

        self.connect(self._address)
    # end connect

    def validate_connected(self) -> None:
        """Validates a connection."""

        self.validate_connection()

        if not self.connected:
            if self.preconnected:
                self.reconnect()

            else:
                raise ValueError(
                    "Socket address is not defined "
                    "for automatic connection."
                )
            # end if
        # end if
    # end validate_connection

    def send(
            self,
            data: bytes,
            connection: Connection = None,
            address: Address = None
    ) -> Output:
        """
        Sends a message to the client or server by its connection.

        :param connection: The sockets' connection object.
        :param data: The message to send to the client.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        self.validate_connected()

        return self.protocol.send(
            connection=connection or self.connection,
            data=data, address=address or self._address
        )
    # end send

    def receive(
            self,
            connection: Connection = None,
            address: Address = None
    ) -> Output:
        """
        Receive a message from the client or server by its connection.

        :param connection: The sockets' connection object.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        self.validate_connected()

        return self.protocol.receive(
            connection=connection or self.connection,
            address=address or self._address
        )
    # end receive

    def close(self) -> None:
        """Closes the connection."""

        super().close()

        self._connected = False
    # end close

    def clone(self) -> Self:
        """
        Returns a copy of the socket wrapper object.

        :return: The new socket object.
        """

        return Client(
            protocol=self.protocol,
            connection=self.protocol.socket(),
            address=self.address,
            reusable=self.reusable
        )
    # end clone
# end Client


# utils.py

import socket
import errno

__all__ = [
    "find_available_port",
    "is_used_port"
]

def find_available_port(host: str) -> int:
    """
    Finds an available port for the host.

    :param host: The host to connect a port with.

    :return: The found port.
    """

    sock = socket.socket()
    sock.bind((host, 0))

    return sock.getsockname()[1]
# end find_available_port

def is_used_port(host: str, port: int) -> bool:
    """
    Checks if a port is already used.

    :param host: The host.
    :param port: The port.

    :return: The boolean flag.
    """

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind((host, port))

        except socket.error as e:
            if e.errno == errno.EADDRINUSE:
                return True

            else:
                raise e
            # end if
        # end try
    # end socket

    return False
# end is_used_port


# setup.py

import codecs

with codecs.open('build.py', 'r') as build_file:
    build_source = build_file.read()

source = dict()

exec(build_source, source)

setup = source['setup']

def main() -> None:
    """Runs the function to distribute the package."""

    setup(
        package="socketsio",
        exclude=[
            "__pycache__",
            "*.pyc"
        ],
        requirements="requirements.txt",
        name='pysocketsio',
        version='1.9.3',
        description=(
            "This module provides a wrapper for the built-in "
            "socket module in python. The program provides server and. "
            "client classes, with the communication methods."
        ),
        license='MIT',
        author="Shahaf Frank-Shapir",
        author_email='shahaffrs@gmail.com',
        long_description_content_type='text/markdown',
        url='https://github.com/Shahaf-F-S/socketsio',
        classifiers=[
            "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Operating System :: OS Independent"
        ]
    )
# end main

if __name__ == "__main__":
    main()
# end if


# protocols.py

import socket
from typing import Callable
from abc import ABCMeta, abstractmethod

from represent import represent

Connection = socket.socket
Address = tuple[str, int]
Output = tuple[bytes, Address | None]

__all__ = [
    "BufferedProtocol",
    "BaseProtocol",
    "EmptyProtocol",
    "TCP",
    "UDP",
    "BCP",
    "BHP",
    "tcp_socket",
    "udp_socket",
    "tcp_bluetooth_socket",
    "is_udp",
    "is_tcp",
    "is_tcp_bluetooth",
    "is_udp_bluetooth",
    "default_protocol",
    "set_default_protocol",
    "reset_default_protocol"
]

def tcp_socket() -> Connection:
    """
    Returns a new TCP socket object.

    :return: The socket object.
    """

    return socket.socket(
        socket.AF_INET,
        socket.SOCK_STREAM
    )
# end tcp_socket

def udp_socket() -> Connection:
    """
    Returns a new UDP socket object.

    :return: The socket object.
    """

    return socket.socket(
        socket.AF_INET,
        socket.SOCK_DGRAM
    )
# end udp_socket

def tcp_bluetooth_socket() -> socket.socket:
    """
    Sends a request through the bluetooth sockets.

    :return: The client object.
    """

    return socket.socket(
        socket.AF_BLUETOOTH,
        socket.SOCK_STREAM,
        socket.BTPROTO_RFCOMM
    )
# end tcp_bluetooth_socket

def udp_bluetooth_socket() -> socket.socket:
    """
    Sends a request through the bluetooth sockets.

    :return: The client object.
    """

    return socket.socket(
        socket.AF_BLUETOOTH,
        socket.SOCK_DGRAM,
        socket.BTPROTO_RFCOMM
    )
# end udp_bluetooth_socket

ProtocolConstructor = Callable[[], "BaseProtocol"] | "BaseProtocol"

@represent
class BaseProtocol(metaclass=ABCMeta):
    """Defines the basic parameters for the communication."""

    DEFAULT: ProtocolConstructor = None

    @classmethod
    def protocol(cls) -> "BaseProtocol":
        """
        Returns a protocol object.

        :return: The default protocol object.
        """

        if cls.DEFAULT is None:
            return BHP()
        # end if

        return cls.DEFAULT()
    # end protocol

    @staticmethod
    @abstractmethod
    def socket() -> Connection:
        """
        Returns a new base socket object.

        :return: The socket object.
        """
    # end socket

    @abstractmethod
    def send(
            self,
            connection: Connection,
            data: bytes,
            address: Address = None
    ) -> Output:
        """
        Sends a message to the client or server by its connection.

        :param data: The message to send to the client.
        :param connection: The sockets' connection object.
        :param address: The address of the sender.

        :return: The received message from the server.
        """
    # end send

    @abstractmethod
    def receive(
            self,
            connection: Connection,
            buffer: int = None,
            address: Address = None
    ) -> Output:
        """
        Receive a message from the client or server by its connection.

        :param connection: The sockets' connection object.
        :param buffer: The buffer size to collect.
        :param address: The address of the sender.

        :return: The received message from the server.
        """
    # end receive
# end BaseProtocol

class BufferedProtocol(BaseProtocol, metaclass=ABCMeta):
    """Defines the basic parameters for the communication."""

    BUFFER = 1024

    def __init__(self, buffer: int = None) -> None:
        """
        Defines the attributes of the protocol.

        :param buffer: The buffer size.
        """

        if buffer is None:
            buffer = self.BUFFER
        # end if

        self.buffer = buffer
    # end __init__
# end BufferedProtocol

class EmptyProtocol(BaseProtocol, metaclass=ABCMeta):
    """Defines the basic parameters for the communication."""

    def __init__(
            self,
            send: Callable[[Connection, bytes, Address | None], Output],
            receive: Callable[[Connection, int | None, Address | None], Output]
    ) -> None:
        """
        Defines the attributes of the protocol.

        :param send: The function to send data.
        :param receive: The function to received data.
        """

        self._send = send
        self._receive = receive
    # end __init__

    def send(
            self,
            connection: Connection,
            data: bytes,
            address: Address = None
    ) -> Output:
        """
        Sends a message to the client or server by its connection.

        :param data: The message to send to the client.
        :param connection: The sockets' connection object.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        return self._send(connection, data, address)
    # end send

    def receive(
            self,
            connection: Connection,
            buffer: int = None,
            address: Address = None
    ) -> Output:
        """
        Receive a message from the client or server by its connection.

        :param connection: The sockets' connection object.
        :param buffer: The buffer size to collect.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        return self._receive(connection, buffer, address)
    # end receive
# end EmptyProtocol

class TCP(BufferedProtocol):
    """Defines the basic parameters for the communication."""

    @staticmethod
    def socket() -> Connection:
        """
        Returns a new base socket object.

        :return: The socket object.
        """

        return tcp_socket()
    # end socket

    def send(
            self,
            connection: Connection,
            data: bytes,
            address: Address = None
    ) -> Output:
        """
        Sends a message to the client or server by its connection.

        :param data: The message to send to the client.
        :param connection: The sockets' connection object.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        connection.send(data)

        return data, address
    # end send

    def receive(
            self,
            connection: Connection,
            buffer: int = None,
            address: Address = None
    ) -> Output:
        """
        Receive a message from the client or server by its connection.

        :param connection: The sockets' connection object.
        :param buffer: The buffer size to collect.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        return connection.recv(buffer or self.buffer), address
    # end receive
# end TCP

class UDP(BufferedProtocol):
    """Defines the basic parameters for the communication."""

    @staticmethod
    def socket() -> Connection:
        """
        Returns a new base socket object.

        :return: The socket object.
        """

        return udp_socket()
    # end socket

    def send(
            self,
            connection: Connection,
            data: bytes,
            address: Address = None
    ) -> Output:
        """
        Sends a message to the client or server by its connection.

        :param data: The message to send to the client.
        :param connection: The sockets' connection object.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        if address is None:
            raise ValueError("address must be a tuple of ip and port.")
        # end if

        connection.sendto(data, address)

        return data, address
    # end send

    def receive(
            self,
            connection: Connection,
            buffer: int = None,
            address: Address = None
    ) -> Output:
        """
        Receive a message from the client or server by its connection.

        :param connection: The sockets' connection object.
        :param buffer: The buffer size to collect.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        return connection.recvfrom(buffer or self.buffer)
    # end receive
# end UDP

class BHP(BaseProtocol):
    """Defines the basic parameters for the communication."""

    HEADER = 32

    def __init__(self, protocol: TCP = None) -> None:
        """
        Defines the base protocol.

        :param protocol: The base protocol object to use.
        """

        self.protocol = protocol or TCP()
    # end __init__

    def socket(self) -> Connection:
        """
        Returns a new base socket object.

        :return: The socket object.
        """

        return self.protocol.socket()
    # end socket

    def send(
            self,
            connection: Connection,
            data: bytes,
            address: Address = None
    ) -> Output:
        """
        Sends a message to the client or server by its connection.

        :param data: The message to send to the client.
        :param connection: The sockets' connection object.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        length_message = str(len(data))
        padding = "0" * (self.HEADER - len(length_message))

        data = (padding + length_message).encode() + data

        return self.protocol.send(
            connection=connection,
            data=data,
            address=address
        )
    # end send

    def receive(
            self,
            connection: Connection,
            buffer: int = None,
            address: Address = None
    ) -> Output:
        """
        Receive a message from the client or server by its connection.

        :param connection: The sockets' connection object.
        :param buffer: The buffer size to collect.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        if buffer is None:
            message, address = self.protocol.receive(
                connection=connection,
                buffer=self.HEADER,
                address=address
            )
            length_message = message.decode()

            if (
                not length_message or
                (length_message.count("0") == len(length_message))
            ):
                return b'', address
            # end if

            buffer = int(length_message)
        # end if

        return self.protocol.receive(
            connection=connection,
            buffer=buffer,
            address=address
        )
    # end receive
# end BCP

class BCP(BHP):
    """Defines the basic parameters for the communication."""

    def __init__(self, protocol: TCP, buffer: int = None) -> None:
        """
        Defines the base protocol.

        :param protocol: The base protocol object to use.
        :param buffer: The buffer size.
        """

        super().__init__(protocol=protocol)

        self._buffer = buffer
    # end __init__

    @property
    def buffer(self) -> int:
        """
        Returns the buffer buffer.

        :return: The buffer buffer.
        """

        if self._buffer is None:
            if not isinstance(self.protocol, BufferedProtocol):
                raise ValueError(
                    f"Size is not defined and protocol is "
                    f"not a subclass of {BufferedProtocol}"
                )
            # end if

            return self.protocol.buffer
        # end if

        return self._buffer
    # end buffer

    @buffer.setter
    def buffer(self, value: int) -> None:
        """
        Returns the buffer buffer.

        :param value: The buffer buffer.
        """

        self._buffer = value
    # end buffer

    def receive(
            self,
            connection: Connection,
            buffer: int = None,
            length: int = None,
            address: Address = None
    ) -> Output:
        """
        Receive a message from the client or server by its connection.

        :param connection: The sockets' connection object.
        :param buffer: The buffer size to collect.
        :param length: The length of the message to expect.
        :param address: The address of the sender.

        :return: The received message from the server.
        """

        if buffer is None:
            message, address = self.protocol.receive(
                connection=connection,
                buffer=self.HEADER,
                address=address
            )
            length_message = message.decode()

            if (
                not length_message or
                (length_message.count("0") == len(length_message))
            ):
                return b'', address
            # end if

            buffer = int(length_message)
        # end if

        buffer = buffer or self.buffer

        if buffer >= length:
            return self.protocol.receive(
                connection=connection,
                buffer=length,
                address=address
            )
        # end if

        data: list[bytes] = []

        for _ in range(length // buffer):
            payload, address = self.protocol.receive(
                connection=connection,
                buffer=buffer,
                address=address
            )

            data.append(payload)
        # end for

        if length % buffer:
            payload, address = self.protocol.receive(
                connection=connection,
                buffer=length % buffer,
                address=address
            )

            data.append(payload)
        # end if

        return b''.join(data), address
    # end receive
# end BCP

def is_tcp(connection: Connection) -> bool:
    """
    Checks if the socket is a TCP socket.

    :param connection: The socket connection object.

    :return: The boolean flag.
    """

    return connection.type == socket.SOCK_STREAM
# end is_tcp

def is_udp(connection: Connection) -> bool:
    """
    Checks if the socket is a UDP socket.

    :param connection: The socket connection object.

    :return: The boolean flag.
    """

    return connection.type == socket.SOCK_DGRAM
# end is_udp

def is_tcp_bluetooth(connection: Connection) -> bool:
    """
    Checks if the socket is a TCP bluetooth socket.

    :param connection: The socket connection object.

    :return: The boolean flag.
    """

    return is_tcp(connection) and (connection.proto == socket.BTPROTO_RFCOMM)
# end is_tcp_bluetooth

def is_udp_bluetooth(connection: Connection) -> bool:
    """
    Checks if the socket is a UDP bluetooth socket.

    :param connection: The socket connection object.

    :return: The boolean flag.
    """

    return is_udp(connection) and (connection.proto == socket.BTPROTO_RFCOMM)
# end is_udp_bluetooth

def default_protocol() -> BaseProtocol:
    """
    Returns a protocol object.

    :return: The default protocol object.
    """

    return BaseProtocol.protocol()
# end default_protocol

def set_default_protocol(constructor: ProtocolConstructor) -> None:
    """
    Sets a new default protocol.

    :param constructor: The default protocol object constructor.
    """

    BaseProtocol.DEFAULT = constructor
# end default_protocol

def reset_default_protocol() -> None:
    """Resets protocol object."""

    BaseProtocol.DEFAULT = None
# end default_protocol